import os
import uuid
from pathlib import Path
from typing import List

from ..conftest import CLOUD_DESTINATION_PREFIXES, get_tested_archive_types
from .resources import CopyTestConfig, DataTestResource


def generate_local_to_cloud_copy_configs() -> List[CopyTestConfig]:
    """Generate test configs for local to cloud data copy

    This method is used to allow for selective test generation,
    instead of generating all possible file/flag combinations as the
    number of tests grows exponentially with the addition of new
    supported platforms
    """
    test_configs: List[CopyTestConfig] = []

    assets_root = (Path(__file__).parent.parent.parent / "assets" / "data").resolve()
    archive_types = get_tested_archive_types()
    local_archives = [
        DataTestResource(
            "local",
            str(assets_root / f"file{ext}"),
            is_archive=True,
            file_extension=ext,
        )
        for ext in archive_types
    ]
    local_folder = DataTestResource(
        "local", str(assets_root) + os.sep, is_archive=False, file_extension=None
    )

    run_uuid = uuid.uuid4().hex

    for schema, prefix in CLOUD_DESTINATION_PREFIXES.items():
        # tests for copy of local folder
        cloud_folder = DataTestResource(
            schema=schema,
            url=f"{prefix}/{run_uuid}/copy/",
            is_archive=False,
            file_extension=None,
        )
        test_configs.append(
            CopyTestConfig(source=local_folder, destination=cloud_folder)
        )

        # tests for compression of local folder
        for archive_type in archive_types:
            cloud_archive = DataTestResource(
                schema=schema,
                url=f"{prefix}/{run_uuid}/compress/file{archive_type}",
                is_archive=True,
                file_extension=archive_type,
            )
            # gzip does not work properly with folders
            # tar should be used instead
            compression_should_fail = archive_type == ".gz"
            compression_fail_reason = "gzip does not support folders properly"
            test_configs.append(
                CopyTestConfig(
                    source=local_folder,
                    destination=cloud_archive,
                    compress_flag=True,
                    should_fail=compression_should_fail,
                    fail_reason=compression_fail_reason,
                )
            )

        # tests for copy of local files
        for archive in local_archives:
            # test for extraction of archive
            cloud_extraction_folder = DataTestResource(
                schema=schema,
                url=f"{prefix}/{run_uuid}/extract/{archive.file_extension}/",
                is_archive=False,
                file_extension=None,
            )
            test_configs.append(
                CopyTestConfig(
                    source=archive,
                    destination=cloud_extraction_folder,
                    extract_flag=True,
                )
            )

            # test for file copy
            cloud_file = DataTestResource(
                schema=schema,
                url=f"{prefix}/{run_uuid}/copy/file{archive.file_extension}",
                is_archive=True,
                file_extension=archive.file_extension,
            )

            # test for skipping compression
            test_configs.append(
                CopyTestConfig(
                    source=archive, destination=cloud_file, compress_flag=True
                )
            )
        if local_archives:
            # use the values for source and destination files
            # from last loop iteration
            test_configs.append(CopyTestConfig(source=archive, destination=cloud_file))

    return test_configs



import time
import typing as t
from contextlib import ExitStack
from decimal import Decimal
from unittest import mock

import neuro_sdk
import pytest
from jose import jwt
from neuro_sdk._config import _AuthConfig, _AuthToken, _ConfigData
from yarl import URL

from neuro_extras.image_builder import ImageBuilder


def _get_mock_presets() -> t.Dict[str, neuro_sdk.Preset]:
    return {
        "cpu-small": neuro_sdk.Preset(
            credits_per_hour=Decimal(1), cpu=1, memory=1 * 1024 * 1024
        ),
        "custom-preset": neuro_sdk.Preset(
            credits_per_hour=Decimal(1),
            cpu=5,
            memory=3 * 1024 * 1024,
            gpu=1,
            gpu_model="mygpu",
        ),
    }


def _get_mock_clusters() -> t.Dict[str, neuro_sdk.Cluster]:
    return {
        "mycluster": neuro_sdk.Cluster(
            name="mycluster",
            registry_url=URL("https://registry.mycluster.noexists"),
            storage_url=URL("https://mycluster.noexists/api/v1/storage"),
            users_url=URL("https://noexists/api/v1/users"),
            monitoring_url=URL("https://mycluster.noexists/api/v1/jobs"),
            secrets_url=URL("https://mycluster.noexists/api/v1/secrets"),
            disks_url=URL("https://mycluster.noexists/api/v1/disks"),
            buckets_url=URL("https://mycluster.noexists/api/v1/buckets"),
            presets=_get_mock_presets(),
            orgs=[],
        ),
    }


def _get_mock_projects() -> t.Dict[neuro_sdk.Project.Key, neuro_sdk.Project]:
    return {
        neuro_sdk.Project.Key(
            cluster_name="mycluster", org_name=None, project_name="myproject"
        ): neuro_sdk.Project(
            cluster_name="mycluster", org_name=None, name="myproject", role="admin"
        ),
    }


def _load_mock_sdk_config() -> _ConfigData:
    return _ConfigData(
        _AuthConfig(
            auth_url=URL("https://notexists/login"),
            token_url=URL("https://notexists/token"),
            logout_url=URL("https://notexists/logout"),
            client_id="myclientid",
            audience="myaudience",
            headless_callback_url=URL("https://notexists/callback"),
        ),
        auth_token=_AuthToken(
            token=jwt.encode({"identity": "myusername"}, "secret"),
            expiration_time=time.time() + 1000,
            refresh_token="myrefreshtoken",
        ),
        url=URL("https://notexists/v1/api"),
        admin_url=URL("https://notexists/api/v1/admin"),
        version="1.0.0",
        project_name="myproject",
        cluster_name="mycluster",
        org_name=None,
        clusters=_get_mock_clusters(),
        projects=_get_mock_projects(),
    )


class MockedNeuroConfig(neuro_sdk.Config):
    def _load(self) -> _ConfigData:
        ret = self.__config_data = _load_mock_sdk_config()
        return ret

    async def check_server(self) -> None:
        pass


@pytest.fixture
async def _neuro_client() -> t.AsyncGenerator[neuro_sdk.Client, None]:
    with ExitStack() as stack:
        stack.enter_context(mock.patch("neuro_sdk.Config", MockedNeuroConfig))
        stack.enter_context(mock.patch("neuro_sdk._client.Config", MockedNeuroConfig))

        stack.enter_context(
            mock.patch("neuro_sdk._storage.Storage.mkdir", mock.AsyncMock())
        )
        stack.enter_context(
            mock.patch("neuro_sdk._storage.Storage.create", mock.AsyncMock())
        )
        stack.enter_context(
            mock.patch("neuro_extras.image._check_image_exists", return_value=False)
        )
        stack.enter_context(mock.patch("uuid.uuid4", return_value="mocked-uuid-4"))
        client = await neuro_sdk.get()
        try:
            yield await client.__aenter__()
        finally:
            await client.__aexit__()


@pytest.fixture
def remote_image_builder(_neuro_client: neuro_sdk.Client) -> ImageBuilder:
    builder_class = ImageBuilder.get(local=False)
    builder_class._execute_subprocess = mock.AsyncMock(  # type: ignore
        side_effect=lambda x: 0
    )

    return builder_class(client=_neuro_client)



from pathlib import Path
from unittest import mock

import pytest
from yarl import URL

from neuro_extras.image import _build_image
from neuro_extras.image_builder import ImageBuilder


async def test_image_builder__min_parameters(
    remote_image_builder: ImageBuilder,
) -> None:
    context = "/path/to/context"

    await _build_image(
        dockerfile_path=Path("path/to/Dockerfile"),
        context=context,
        image_uri_str="image:targetimage:latest",
        use_cache=True,
        build_args=(),
        volume=(),
        env=(),
        build_tags=(),
        force_overwrite=False,
    )

    expected_storage_build_root = URL(
        "storage://mycluster/myproject/.builds/mocked-uuid-4"
    )
    storage_mkdir_mock: mock.AsyncMock = remote_image_builder._client.storage.mkdir  # type: ignore # noqa: E501
    storage_mkdir_mock.assert_awaited_once_with(
        expected_storage_build_root, parents=True
    )
    storage_create_mock: mock.AsyncMock = remote_image_builder._client.storage.create  # type: ignore # noqa: E501
    storage_create_mock.assert_awaited_once_with(
        expected_storage_build_root / ".docker.config.json",
        mock.ANY,
    )
    subproc_mock: mock.AsyncMock = remote_image_builder._execute_subprocess  # type: ignore # noqa: E501
    assert subproc_mock.await_count == 2
    upload_ctx_cmd = subproc_mock.await_args_list[0][0][0]
    assert upload_ctx_cmd == [
        "neuro",
        "--disable-pypi-version-check",
        "cp",
        "--recursive",
        Path(context).resolve().as_uri(),
        str(expected_storage_build_root / "context"),
    ]
    start_build_cmd = subproc_mock.await_args_list[1][0][0]
    start_build_neuro_args = start_build_cmd[: start_build_cmd.index("--")]
    start_build_job_arg = start_build_cmd[start_build_cmd.index("--") + 1 :][0]
    start_build_kaniko_args = start_build_job_arg.split(" ")
    assert start_build_neuro_args == [
        "neuro",
        "--disable-pypi-version-check",
        "job",
        "run",
        "--life-span=4h",
        "--schedule-timeout=20m",
        "--project=myproject",
        "--tag=kaniko-builds-image:image://mycluster/myproject/targetimage:latest",
        "--volume=storage://mycluster/myproject/.builds/mocked-uuid-4/.docker.config.json:/kaniko/.docker/config.json:ro",  # noqa: E501
        "--volume=storage://mycluster/myproject/.builds/mocked-uuid-4/context:/kaniko_context:rw",  # noqa: E501
        "--env=container=docker",
        "gcr.io/kaniko-project/executor:v1.20.0-debug",
    ]
    assert start_build_kaniko_args == [
        "--context=/kaniko_context",
        "--dockerfile=/kaniko_context/path/to/Dockerfile",
        "--destination=registry.mycluster.noexists/myproject/targetimage:latest",
        "--cache=true",
        "--cache-repo=registry.mycluster.noexists/myproject/layer-cache/cache",
        "--verbosity=info",
        "--image-fs-extract-retry=1",
        "--push-retry=3",
        "--use-new-run=true",
        "--snapshot-mode=redo",
    ]


async def test_image_builder__full_parameters(
    remote_image_builder: ImageBuilder,
) -> None:
    context = "/path/to/context"

    await _build_image(
        dockerfile_path=Path("path/to/Dockerfile"),
        context=context,
        image_uri_str="image:targetimage:latest",
        use_cache=True,
        build_args=("ARG1=ARGVAL1", "ARG2=ARGVAL2"),
        volume=(
            "storage:somevol:/mnt/vol1",
            "storage:/someproject2/somevol2:/mnt/vol2",
        ),
        env=("ENV1=VAL1", "ENV2=VAL2"),
        preset="custom-preset",
        build_tags=("tag1", "tag2"),
        project_name="myproject",
        extra_kaniko_args="--some-kaniko-arg1 --some-kaniko-arg2=arg2val",
        force_overwrite=False,
    )

    expected_storage_build_root = URL(
        "storage://mycluster/myproject/.builds/mocked-uuid-4"
    )
    storage_mkdir_mock: mock.AsyncMock = remote_image_builder._client.storage.mkdir  # type: ignore # noqa: E501
    storage_mkdir_mock.assert_awaited_once_with(
        expected_storage_build_root, parents=True
    )
    storage_create_mock: mock.AsyncMock = remote_image_builder._client.storage.create  # type: ignore # noqa: E501
    storage_create_mock.assert_awaited_once_with(
        expected_storage_build_root / ".docker.config.json",
        mock.ANY,
    )
    subproc_mock: mock.AsyncMock = remote_image_builder._execute_subprocess  # type: ignore # noqa: E501
    assert subproc_mock.await_count == 2
    upload_ctx_cmd = subproc_mock.await_args_list[0][0][0]
    assert upload_ctx_cmd == [
        "neuro",
        "--disable-pypi-version-check",
        "cp",
        "--recursive",
        Path(context).resolve().as_uri(),
        str(expected_storage_build_root / "context"),
    ]
    start_build_cmd = subproc_mock.await_args_list[1][0][0]
    start_build_neuro_args = start_build_cmd[: start_build_cmd.index("--")]
    start_build_job_arg = start_build_cmd[start_build_cmd.index("--") + 1 :][0]
    start_build_kaniko_args = start_build_job_arg.split(" ")
    assert start_build_neuro_args == [
        "neuro",
        "--disable-pypi-version-check",
        "job",
        "run",
        "--life-span=4h",
        "--schedule-timeout=20m",
        "--project=myproject",
        "--preset=custom-preset",
        "--tag=tag1",
        "--tag=tag2",
        "--tag=kaniko-builds-image:image://mycluster/myproject/targetimage:latest",
        "--volume=storage:somevol:/mnt/vol1",
        "--volume=storage:/someproject2/somevol2:/mnt/vol2",
        "--volume=storage://mycluster/myproject/.builds/mocked-uuid-4/.docker.config.json:/kaniko/.docker/config.json:ro",  # noqa: E501
        "--volume=storage://mycluster/myproject/.builds/mocked-uuid-4/context:/kaniko_context:rw",  # noqa: E501
        "--env=ENV1=VAL1",
        "--env=ENV2=VAL2",
        "--env=container=docker",
        "gcr.io/kaniko-project/executor:v1.20.0-debug",
    ]
    assert start_build_kaniko_args == [
        "--context=/kaniko_context",
        "--dockerfile=/kaniko_context/path/to/Dockerfile",
        "--destination=registry.mycluster.noexists/myproject/targetimage:latest",
        "--cache=true",
        "--cache-repo=registry.mycluster.noexists/myproject/layer-cache/cache",
        "--verbosity=info",
        "--image-fs-extract-retry=1",
        "--push-retry=3",
        "--use-new-run=true",
        "--snapshot-mode=redo",
        "--build-arg",
        "ARG1=ARGVAL1",
        "--build-arg",
        "ARG2=ARGVAL2",
        "--build-arg",
        "ENV1",
        "--build-arg",
        "ENV2",
        "--some-kaniko-arg1",
        "--some-kaniko-arg2=arg2val",
    ]


async def test_image_builder__conflicting_kaniko_args(
    remote_image_builder: ImageBuilder,
) -> None:
    with pytest.raises(
        ValueError,
        match=(
            "Extra kaniko arguments {'--image-fs-extract-retry'} overlap "
            "with autogenerated arguments. Please remove them "
            "in order to proceed or contact the support team."
        ),
    ):
        await _build_image(
            dockerfile_path=Path("path/to/Dockerfile"),
            context="/path/to/context",
            image_uri_str="image:targetimage:latest",
            use_cache=True,
            build_args=(),
            volume=(),
            env=(),
            preset=None,
            build_tags=(),
            project_name="myproject",
            extra_kaniko_args="--image-fs-extract-retry=3",
            force_overwrite=False,
        )


async def test_image_builder__custom_project(
    remote_image_builder: ImageBuilder,
) -> None:
    context = "/path/to/context"
    await _build_image(
        dockerfile_path=Path("path/to/Dockerfile"),
        context=context,
        image_uri_str="image:targetimage:latest",
        use_cache=True,
        build_args=(),
        volume=(),
        env=(),
        build_tags=(),
        force_overwrite=False,
        project_name="otherproject",
    )

    expected_storage_build_root = URL(
        "storage://mycluster/otherproject/.builds/mocked-uuid-4"
    )
    storage_mkdir_mock: mock.AsyncMock = remote_image_builder._client.storage.mkdir  # type: ignore # noqa: E501
    storage_mkdir_mock.assert_awaited_once_with(
        expected_storage_build_root, parents=True
    )
    storage_create_mock: mock.AsyncMock = remote_image_builder._client.storage.create  # type: ignore # noqa: E501
    storage_create_mock.assert_awaited_once_with(
        expected_storage_build_root / ".docker.config.json",
        mock.ANY,
    )
    subproc_mock: mock.AsyncMock = remote_image_builder._execute_subprocess  # type: ignore # noqa: E501
    assert subproc_mock.await_count == 2
    upload_ctx_cmd = subproc_mock.await_args_list[0][0][0]
    assert upload_ctx_cmd == [
        "neuro",
        "--disable-pypi-version-check",
        "cp",
        "--recursive",
        Path(context).resolve().as_uri(),
        str(expected_storage_build_root / "context"),
    ]
    start_build_cmd = subproc_mock.await_args_list[1][0][0]
    start_build_neuro_args = start_build_cmd[: start_build_cmd.index("--")]
    start_build_job_arg = start_build_cmd[start_build_cmd.index("--") + 1 :][0]
    start_build_kaniko_args = start_build_job_arg.split(" ")
    assert start_build_neuro_args == [
        "neuro",
        "--disable-pypi-version-check",
        "job",
        "run",
        "--life-span=4h",
        "--schedule-timeout=20m",
        "--project=otherproject",
        "--tag=kaniko-builds-image:image://mycluster/otherproject/targetimage:latest",
        "--volume=storage://mycluster/otherproject/.builds/mocked-uuid-4/.docker.config.json:/kaniko/.docker/config.json:ro",  # noqa: E501
        "--volume=storage://mycluster/otherproject/.builds/mocked-uuid-4/context:/kaniko_context:rw",  # noqa: E501
        "--env=container=docker",
        "gcr.io/kaniko-project/executor:v1.20.0-debug",
    ]
    assert start_build_kaniko_args == [
        "--context=/kaniko_context",
        "--dockerfile=/kaniko_context/path/to/Dockerfile",
        "--destination=registry.mycluster.noexists/otherproject/targetimage:latest",
        "--cache=true",
        "--cache-repo=registry.mycluster.noexists/otherproject/layer-cache/cache",
        "--verbosity=info",
        "--image-fs-extract-retry=1",
        "--push-retry=3",
        "--use-new-run=true",
        "--snapshot-mode=redo",
    ]


async def test_image_builder__storage_context(
    remote_image_builder: ImageBuilder,
) -> None:
    context_uri_str = "storage:context"
    await _build_image(
        dockerfile_path=Path("path/to/Dockerfile"),
        context=context_uri_str,
        image_uri_str="image:targetimage:latest",
        use_cache=True,
        build_args=(),
        volume=(),
        env=(),
        build_tags=(),
        force_overwrite=False,
    )

    expected_storage_build_root = URL(
        "storage://mycluster/myproject/.builds/mocked-uuid-4"
    )
    storage_mkdir_mock: mock.AsyncMock = remote_image_builder._client.storage.mkdir  # type: ignore # noqa: E501
    storage_mkdir_mock.assert_awaited_once_with(
        expected_storage_build_root, parents=True
    )
    storage_create_mock: mock.AsyncMock = remote_image_builder._client.storage.create  # type: ignore # noqa: E501
    storage_create_mock.assert_awaited_once_with(
        expected_storage_build_root / ".docker.config.json",
        mock.ANY,
    )
    subproc_mock: mock.AsyncMock = remote_image_builder._execute_subprocess  # type: ignore # noqa: E501
    assert subproc_mock.await_count == 1
    start_build_cmd = subproc_mock.await_args_list[0][0][0]
    start_build_neuro_args = start_build_cmd[: start_build_cmd.index("--")]
    start_build_job_arg = start_build_cmd[start_build_cmd.index("--") + 1 :][0]
    start_build_kaniko_args = start_build_job_arg.split(" ")
    assert start_build_neuro_args == [
        "neuro",
        "--disable-pypi-version-check",
        "job",
        "run",
        "--life-span=4h",
        "--schedule-timeout=20m",
        "--project=myproject",
        "--tag=kaniko-builds-image:image://mycluster/myproject/targetimage:latest",
        "--volume=storage://mycluster/myproject/.builds/mocked-uuid-4/.docker.config.json:/kaniko/.docker/config.json:ro",  # noqa: E501
        "--volume=storage://mycluster/myproject/context:/kaniko_context:rw",
        "--env=container=docker",
        "gcr.io/kaniko-project/executor:v1.20.0-debug",
    ]
    assert start_build_kaniko_args == [
        "--context=/kaniko_context",
        "--dockerfile=/kaniko_context/path/to/Dockerfile",
        "--destination=registry.mycluster.noexists/myproject/targetimage:latest",
        "--cache=true",
        "--cache-repo=registry.mycluster.noexists/myproject/layer-cache/cache",
        "--verbosity=info",
        "--image-fs-extract-retry=1",
        "--push-retry=3",
        "--use-new-run=true",
        "--snapshot-mode=redo",
    ]



"""Module for copying files from HTTP(S) sources"""
from ..utils import CLIRunner
from .common import Copier, DataUrlType, Resource


class WebCopier(Copier, CLIRunner):
    """Copier for downloading data from HTTP(S) sources"""

    def _ensure_can_execute(self) -> None:
        if not (
            self.source.data_url_type == DataUrlType.WEB
            and self.destination.data_url_type == DataUrlType.LOCAL_FS
        ):
            raise ValueError(
                f"Can only copy from {DataUrlType.WEB.name} "
                f"to {DataUrlType.LOCAL_FS.name}"
            )

    async def perform_copy(self) -> Resource:
        """Perform copy through running rclone and return the url to destinaton"""
        if not self.source.data_url_type == DataUrlType.WEB:
            raise ValueError("Only copy from HTTP(s) sources is supported")
        if self.destination.data_url_type == DataUrlType.WEB:
            raise ValueError("Copy to HTTP(S) destinations is unsupported")
        if self.source.filename is None:
            raise ValueError(
                "Copy from HTTP(S) directory is unsupported. "
                "Please, reach us at https://github.com/neuro-inc/neuro-extras/issues "
                "describing your use case."
            )
        command = "rclone"
        args = [
            "copyto",
            "--http-url",
            # HTTP URL parameter for rclone is just scheme + host name
            str(self.source.url.with_path("").with_query("")),
            f":http:{self.source.url.path}",
            self.destination.as_str(),
        ]
        await self.run_command(command=command, args=args)
        return self.destination



import dataclasses
import logging
import re
from pathlib import Path
from time import sleep
from typing import List, Optional

from neuro_sdk import Client
from yarl import URL

from neuro_extras.data.azure import _build_sas_url, _patch_azure_url_for_rclone
from neuro_extras.data.common import Resource

from ..conftest import DISK_PREFIX, TEMPDIR_PREFIX
from .utils import _run_command


logger = logging.getLogger(__name__)


@dataclasses.dataclass
class DataTestResource:
    schema: str
    url: str
    is_archive: bool
    file_extension: Optional[str] = None
    client: Optional[Client] = None

    def remove(self) -> None:
        """Remove the resource at self.url"""
        removal_handlers = {
            "local": _remove_local,
            "s3": _remove_s3,
            "azure+https": _remove_azure,
            "gs": _remove_gcs,
            "storage": _remove_storage,
        }
        if self.schema in removal_handlers:
            logger.info(f"Removing {self.url}")
            removal_handlers[self.schema](self.url)
        else:
            logger.warning(f"No cleanup handler for schema '{self.schema}'")

    def exists(self) -> bool:
        """Check if the resource exists and is of appropriate type (dir or file)"""
        handlers = {
            "local": _local_resource_exists,
            "s3": _s3_resource_exists,
            "gs": _gcs_resource_exists,
            "azure+https": _azure_resource_exist,
            "storage": _storage_resource_exists,
            "disk": _resource_on_disk_exists,
        }
        if self.schema in handlers:
            logger.info(f"Checking if {self.url} exists")
            return handlers[self.schema](self)
        else:
            logger.error(f"Check for '{self.schema}' schema is not implemented!")
            return False

    def patch_disk(self, disk: str) -> None:
        """Patch disk url containing DISK_PREFIX with specified disk id"""
        if self.schema == "disk":
            patched = re.sub(DISK_PREFIX, disk, self.url)
            logger.info(f"Patched {self.url} into {patched}")
            self.url = patched

    def patch_tempdir(self, dir: str) -> None:
        if self.schema == "local":
            patched = re.sub(TEMPDIR_PREFIX, dir, self.url)
            logger.info(f"Patched {self.url} into {patched}")
            self.url = patched


def _remove_local(url: str) -> None:
    returncode, stdout, stderr = _run_command("rm", ["-rvf", url])
    assert returncode == 0, stderr


def _remove_gcs(url: str) -> None:
    returncode, stdout, stderr = _run_command("gsutil", ["-m", "rm", "-r", url])
    assert returncode == 0, stderr


def _remove_s3(url: str) -> None:
    returncode, stdout, stderr = _run_command("aws", ["s3", "rm", "--recursive", url])
    assert returncode == 0, stderr


def _remove_storage(url: str) -> None:
    returncode, stdout, stderr = _run_command("neuro", ["storage", "rm", "-r", url])
    assert returncode == 0 or "Not Found" in stderr


def _remove_azure(url: str) -> None:
    azure_sas_url = _build_sas_url(URL(url))
    patched_url = _patch_azure_url_for_rclone(URL(url))
    returncode, stdout, stderr = _run_command(
        "rclone", ["delete", "--azureblob-sas-url", azure_sas_url, patched_url]
    )
    assert returncode == 0, stderr


def _local_resource_exists(resource: DataTestResource) -> bool:
    path = Path(resource.url)
    path_exists = path.exists()
    valid_type = (
        path.is_file() if resource.file_extension is not None else path.is_dir()
    )
    return path_exists and valid_type


def _s3_resource_exists(resource: DataTestResource) -> bool:
    if resource.file_extension:
        url = URL(resource.url)
        bucket_name = url.host
        filename = "/".join(url.parts[1:])
        command = "aws"
        args = ["s3api", "head-object", "--bucket", bucket_name, "--key", filename]
    else:
        command = "aws"
        args = [
            "s3",
            "ls",
            resource.url,
        ]
    returcode, stdout, stederr = _run_command(command, args)  # type: ignore
    return returcode == 0


def _storage_resource_exists(resource: DataTestResource) -> bool:
    sleep(5)  # (A.K.) I hate to admit it, but without this it often does not work
    command = "neuro"
    assert resource.client is not None
    if resource.file_extension is None:
        args = [
            "storage",
            "ls",
            "-l",
            Resource.parse(resource.url, client=resource.client).as_str(),
        ]
    else:
        args = [
            "storage",
            "ls",
            "-l",
            Resource.parse(resource.url, client=resource.client)
            .strip_filename()
            .as_str(),
        ]
    returncode, stdout, stderr = _run_command(command, args)
    if resource.file_extension is None:
        check_is_successful = True
    else:
        stdout = stdout if stdout else ""
        filename = Resource.parse(resource.url, client=resource.client).filename
        assert filename is not None
        check_is_successful = filename in stdout
    return returncode == 0 and check_is_successful


def _resource_on_disk_exists(resource: DataTestResource) -> bool:
    command = "neuro"
    assert resource.client
    disk_id, path_on_disk = Resource.parse(
        resource.url, client=resource.client
    ).disk_id_and_path
    mountpoint = "/var/mnt"
    path_in_job = f"{mountpoint}{path_on_disk if path_on_disk else '/'}"
    args = [
        "job",
        "run",
        "-v",
        f"{disk_id}:{mountpoint}",
        "busybox",
        "--",
        "stat",
        path_in_job,
    ]
    returncode, stdout, stderr = _run_command(command, args)
    return returncode == 0


def _gcs_resource_exists(resource: DataTestResource) -> bool:
    command = "gsutil"
    args = [
        "ls",
        resource.url,
    ]
    returncode, _, _ = _run_command(command, args)
    return returncode == 0


def _azure_resource_exist(resource: DataTestResource) -> bool:
    command = "rclone"
    args = [
        "-q",
        "--azureblob-sas-url",
        _build_sas_url(URL(resource.url)),
        "lsf",
        _patch_azure_url_for_rclone(URL(resource.url)),
    ]
    returncode, stdout, stderr = _run_command(command, args)
    return returncode == 0 and bool(stdout)


@dataclasses.dataclass
class CopyTestConfig:
    source: DataTestResource
    destination: DataTestResource
    extract_flag: bool = False
    compress_flag: bool = False
    # if set, force testcase to be expected to fail
    should_fail: Optional[bool] = None
    # if should_fail is set - provide reason
    fail_reason: Optional[str] = None

    def as_command(self, minimized: bool = False) -> List[str]:
        if minimized:
            command = []
        else:
            command = ["data", "cp"] + self.get_extra_args()
        if self.extract_flag:
            command.append("-x")
        if self.compress_flag:
            command.append("-c")
        if minimized:
            source_filepart = (
                self.source.file_extension if self.source.file_extension else "/"
            )
            source_url = f"{self.source.schema}:***{source_filepart}"
            destination_filepart = (
                self.destination.file_extension
                if self.destination.file_extension
                else "/"
            )
            destination_url = f"{self.destination.schema}:***{destination_filepart}"
        else:
            source_url = self.source.url
            destination_url = self.destination.url
        command += [
            source_url,
            destination_url,
        ]
        return command

    def __str__(self) -> str:
        return " ".join(self.as_command(minimized=True))

    def get_extra_args(self) -> List[str]:
        schemas = (self.source.schema, self.destination.schema)
        if not (("storage" in schemas) or ("disk" in schemas)):
            return []
        extra_args = []
        if "gs" in schemas:
            extra_args += [
                "-v",
                "secret:neuro-extras-gcp:/gcp-creds.txt",
                "-e",
                "GOOGLE_APPLICATION_CREDENTIALS=/gcp-creds.txt",
            ]
        if "azure+https" in schemas:
            extra_args += ["-e", "AZURE_SAS_TOKEN=secret:azure_sas_token"]
        if "s3" in schemas:
            extra_args += [
                "-v",
                "secret:neuro-extras-aws:/aws-creds.txt",
                "-e",
                "AWS_CONFIG_FILE=/aws-creds.txt",
            ]
        extra_args += ["--preset", "cpu-small"]
        return extra_args



from typing import Dict, Iterator

import pytest
from neuro_sdk import Client, Preset


class MockNeuroClient(Client):
    def __init__(self) -> None:
        self._presets: Dict[str, Preset] = {}

    @property
    def presets(self) -> Dict[str, Preset]:
        return self._presets


@pytest.fixture
def mock_client() -> Iterator[MockNeuroClient]:
    yield MockNeuroClient._create()



import sys
import uuid
from pathlib import Path
from unittest import mock

import pytest
import yaml

from neuro_extras.common import NEURO_EXTRAS_IMAGE

from .conftest import CLIRunner


@pytest.mark.smoke
@pytest.mark.skipif(sys.platform == "win32", reason="kaniko does not work on Windows")
def test_seldon_deploy_from_local(cli_runner: CLIRunner) -> None:
    result = cli_runner(["neuro-extras", "init-aliases"])
    assert result.returncode == 0, result

    pkg_path = Path("pkg")
    result = cli_runner(["neuro", "seldon-init-package", str(pkg_path)])
    assert result.returncode == 0, result
    assert "Copying a Seldon package scaffolding" in result.stdout, result

    assert (pkg_path / "seldon.Dockerfile").exists()

    tag = str(uuid.uuid4())
    img_uri = f"image:extras-e2e:{tag}"
    result = cli_runner(
        ["neuro", "image-build", "-f", "seldon.Dockerfile", str(pkg_path), img_uri]
    )
    assert result.returncode == 0, result

    result = cli_runner(["neuro", "image", "size", img_uri])
    assert result.returncode == 0, result

    cli_runner(["neuro", "image", "rm", img_uri])


def test_seldon_generate_deployment(cli_runner: CLIRunner) -> None:
    result = cli_runner(
        [
            "neuro-extras",
            "seldon",
            "generate-deployment",
            "image:model:latest",
            "storage:model/model.pkl",
        ]
    )
    assert result.returncode == 0, result

    payload = yaml.safe_load(result.stdout)
    expected_pod_spec = {
        "volumes": [
            {"emptyDir": {}, "name": "neuro-storage"},
            {"name": "neuro-secret", "secret": {"secretName": "neuro"}},
        ],
        "imagePullSecrets": [{"name": "neuro-registry"}],
        "initContainers": [
            {
                "name": "neuro-download",
                "image": NEURO_EXTRAS_IMAGE,
                "imagePullPolicy": "Always",
                "securityContext": {"runAsUser": 0},
                "command": ["bash", "-c"],
                "args": [
                    "cp -L -r /var/run/neuro/config /root/.neuro;"
                    "chmod 0700 /root/.neuro;"
                    "chmod 0600 /root/.neuro/db;"
                    "neuro cp storage:model/model.pkl /storage"
                ],
                "volumeMounts": [
                    {"mountPath": "/storage", "name": "neuro-storage"},
                    {"mountPath": "/var/run/neuro/config", "name": "neuro-secret"},
                ],
            }
        ],
        "containers": [
            {
                "name": "model",
                "image": mock.ANY,
                "imagePullPolicy": "Always",
                "volumeMounts": [{"mountPath": "/storage", "name": "neuro-storage"}],
            }
        ],
    }
    assert payload == {
        "apiVersion": "machinelearning.seldon.io/v1",
        "kind": "SeldonDeployment",
        "metadata": {"name": "neuro-model"},
        "spec": {
            "predictors": [
                {
                    "componentSpecs": [{"spec": expected_pod_spec}],
                    "graph": {
                        "endpoint": {"type": "REST"},
                        "name": "model",
                        "type": "MODEL",
                    },
                    "name": "predictor",
                    "replicas": 1,
                }
            ]
        },
    }


def test_seldon_generate_deployment_custom(cli_runner: CLIRunner) -> None:
    result = cli_runner(
        [
            "neuro-extras",
            "seldon",
            "generate-deployment",
            "--name",
            "test",
            "--neuro-secret",
            "test-neuro",
            "--registry-secret",
            "test-registry",
            "image:model:latest",
            "storage:model/model.pkl",
        ]
    )
    assert result.returncode == 0, result

    payload = yaml.safe_load(result.stdout)
    expected_pod_spec = {
        "volumes": [
            {"emptyDir": {}, "name": "neuro-storage"},
            {"name": "neuro-secret", "secret": {"secretName": "test-neuro"}},
        ],
        "imagePullSecrets": [{"name": "test-registry"}],
        "initContainers": [
            {
                "name": "neuro-download",
                "image": NEURO_EXTRAS_IMAGE,
                "imagePullPolicy": "Always",
                "securityContext": {"runAsUser": 0},
                "command": ["bash", "-c"],
                "args": [
                    "cp -L -r /var/run/neuro/config /root/.neuro;"
                    "chmod 0700 /root/.neuro;"
                    "chmod 0600 /root/.neuro/db;"
                    "neuro cp storage:model/model.pkl /storage"
                ],
                "volumeMounts": [
                    {"mountPath": "/storage", "name": "neuro-storage"},
                    {"mountPath": "/var/run/neuro/config", "name": "neuro-secret"},
                ],
            }
        ],
        "containers": [
            {
                "name": "model",
                "image": mock.ANY,
                "imagePullPolicy": "Always",
                "volumeMounts": [{"mountPath": "/storage", "name": "neuro-storage"}],
            }
        ],
    }
    assert payload == {
        "apiVersion": "machinelearning.seldon.io/v1",
        "kind": "SeldonDeployment",
        "metadata": {"name": "test"},
        "spec": {
            "predictors": [
                {
                    "componentSpecs": [{"spec": expected_pod_spec}],
                    "graph": {
                        "endpoint": {"type": "REST"},
                        "name": "model",
                        "type": "MODEL",
                    },
                    "name": "predictor",
                    "replicas": 1,
                }
            ]
        },
    }



import logging

import click

from .utils import setup_child_watcher
from .version import __version__


setup_child_watcher()


class ClickLogHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        try:
            msg = self.format(record)
            click.echo(msg)
        except Exception:
            self.handleError(record)


@click.group()
@click.option(
    "-v",
    "--verbose",
    count=True,
    type=int,
    default=0,
    help="Give more output. Option is additive, and can be used up to 2 times.",
)
@click.option(
    "-q",
    "--quiet",
    count=True,
    type=int,
    default=0,
    help="Give less output. Option is additive, and can be used up to 2 times.",
)
@click.version_option(
    version=__version__, message="neuro-extras package version: %(version)s"
)
def main(
    verbose: int,
    quiet: int,
) -> None:
    """
    Auxiliary scripts and recipes for automating routine tasks.
    """
    handler = ClickLogHandler()
    handler.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))

    root_logger = logging.getLogger()
    verbosity = verbose - quiet
    if verbosity < -1:
        loglevel = logging.CRITICAL
    elif verbosity == -1:
        loglevel = logging.WARNING
    elif verbosity == 0:
        loglevel = logging.INFO
    else:
        loglevel = logging.DEBUG
    handler.setLevel(loglevel)

    root_logger.setLevel(logging.DEBUG)
    root_logger.addHandler(handler)



from decimal import Decimal

import pytest
from neuro_sdk import Preset

from neuro_extras.utils import select_job_preset

from .conftest import MockNeuroClient


FAKE_PRESETS = {
    "bad": Preset(cpu=1, memory=9999 * 2**20, credits_per_hour=Decimal("5")),
    "expensive": Preset(cpu=2, memory=9999 * 2**20, credits_per_hour=Decimal("15")),
    "cheap": Preset(cpu=20, memory=99999 * 2**20, credits_per_hour=Decimal("14")),
    "cheap_scheduled": Preset(
        cpu=20,
        memory=99999 * 2**20,
        credits_per_hour=Decimal("10"),
        scheduler_enabled=True,
    ),
}


def test_cheapest_preset_is_selected(mock_client: MockNeuroClient) -> None:
    mock_client.presets.update(FAKE_PRESETS)
    selected_preset = select_job_preset(
        preset=None, client=mock_client, min_mem=4096, min_cpu=2
    )
    assert selected_preset == "cheap"


@pytest.mark.parametrize("preset", ["bad", "cheap_scheduled"])
def test_user_selection_is_respected(mock_client: MockNeuroClient, preset: str) -> None:
    mock_client.presets.update(FAKE_PRESETS)
    selected_preset = select_job_preset(
        preset=preset, client=mock_client, min_mem=4096, min_cpu=2
    )
    assert selected_preset == preset


def test_when_nothing_fits_first_preset_is_used(mock_client: MockNeuroClient) -> None:
    presets = {
        "bad": Preset(cpu=1, memory=9999, credits_per_hour=Decimal("5")),
        "gpu": Preset(cpu=4, memory=9999, credits_per_hour=Decimal("15"), gpu=1),
    }
    selected_preset = select_job_preset(
        preset=None, client=mock_client, min_mem=4096, min_cpu=2
    )
    mock_client.presets.update(presets)
    assert selected_preset is None



"""Module for copying files from/to Google Cloud Storage"""
from ..utils import CLIRunner
from .common import Copier, DataUrlType, Resource


class GCSCopier(Copier, CLIRunner):
    """Copier, that is capable of copying to/from Google Cloud Storage"""

    def _ensure_can_execute(self) -> None:
        if not (
            self.source.data_url_type == DataUrlType.LOCAL_FS
            and self.destination.data_url_type == DataUrlType.GCS
            or self.source.data_url_type == DataUrlType.GCS
            and self.destination.data_url_type == DataUrlType.LOCAL_FS
        ):
            raise ValueError(
                "Unsupported source and destination - "
                f"can only copy between {DataUrlType.GCS.name} "
                f"and {DataUrlType.LOCAL_FS.name}"
            )

    async def perform_copy(self) -> Resource:
        """Perform copy through running gsutil and return the url to destinaton"""

        command = "gsutil"
        args = ["-m", "cp", "-r", str(self.source.url), str(self.destination.url)]
        await self.run_command(command=command, args=args)
        return self.destination



import json
import logging
import os
import subprocess
import sys
from pathlib import Path

import pytest
from deepdiff import DeepDiff


TEST_ASSETS_ROOT = (Path(__file__).parent.parent / "assets").resolve()
NEURO_EXTRAS_ROOT = Path(__file__).parent.parent.parent / "neuro_extras"
LOGGER = logging.getLogger(__name__)


@pytest.mark.skipif(sys.platform != "darwin", reason="Need sh to test this test.")
def test_auth_merge_script(tmp_path: Path) -> None:
    auths = TEST_ASSETS_ROOT / "registry_auths"
    script = NEURO_EXTRAS_ROOT / "assets" / "merge_docker_auths.sh"
    auth1 = auths / "default-registry.json"
    auth2 = auths / "dockerhub-registry.json"
    merged = auths / "merged.json"
    result_file = tmp_path / "result.json"
    # Check assets
    assert script.exists() and script.is_file()
    assert auth1.exists() and auth1.is_file()
    assert auth2.exists() and auth2.is_file()
    assert merged.exists() and merged.is_file()

    os.environ["NE_REGISTRY_AUTH_INVALID"] = "hey there!"
    os.environ["NE_REGISTRY_AUTH_VALID_STR"] = auth1.read_text().replace("\n", "")
    os.environ["NE_REGISTRY_AUTH_VALID_FILE"] = str(auth2)
    os.environ["NE_RESULT_PATH"] = str(result_file)

    proc = subprocess.Popen(
        ["sh", str(script)], stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    stdout_b, stderr_b = proc.communicate(timeout=10)
    stdout, stderr = stdout_b.decode(), stderr_b.decode()
    LOGGER.info(stdout)
    LOGGER.info(stderr)
    assert proc.returncode == 0, f"STDOUT: {stdout}\nSTDERR: {stderr}"
    assert (
        stderr.count("NE_REGISTRY") == 1
    ), "Not only 'NE_REGISTRY_AUTH_INVALID' was asssumed invalid"
    expected = json.loads(merged.read_text())
    actual = json.loads(result_file.read_text())
    diff = DeepDiff(expected, actual, ignore_order=True)

    assert diff == {}, diff



import logging
from typing import List

from ..conftest import CLOUD_SOURCE_PREFIXES, get_tested_archive_types
from .resources import TEMPDIR_PREFIX, CopyTestConfig, DataTestResource


logger = logging.getLogger(__name__)


def generate_cloud_to_local_copy_configs() -> List[CopyTestConfig]:
    test_configs: List[CopyTestConfig] = []
    archive_types = get_tested_archive_types()

    for schema, cloud_url in CLOUD_SOURCE_PREFIXES.items():
        cloud_folder = DataTestResource(
            schema=schema,
            url=f"{cloud_url}/",
            is_archive=False,
            file_extension=None,
        )
        local_copy_folder = DataTestResource(
            schema="local",
            url=f"{TEMPDIR_PREFIX}/copy/{schema}/folder/",
            is_archive=False,
            file_extension=None,
        )
        for ext in archive_types:
            # test for copy of remote archive into local one
            cloud_archive = DataTestResource(
                schema=schema,
                url=f"{cloud_url}/file{ext}",
                is_archive=True,
                file_extension=ext,
            )
            local_archive = DataTestResource(
                schema="local",
                url=f"{TEMPDIR_PREFIX}/copy/{schema}/file{ext}",
                is_archive=True,
                file_extension=ext,
            )

            test_configs.append(
                CopyTestConfig(
                    source=cloud_archive,
                    destination=local_archive,
                    compress_flag=True,
                )
            )
            # test for extraction of cloud archive into local dir
            local_extract_folder = DataTestResource(
                schema="local",
                url=f"{TEMPDIR_PREFIX}/extract/{schema}/{ext}/",
                is_archive=False,
            )
            test_configs.append(
                CopyTestConfig(
                    source=cloud_archive,
                    destination=local_extract_folder,
                    extract_flag=True,
                )
            )

            # test for compression of cloud folder into local archive

            local_compressed_archive = DataTestResource(
                schema="local",
                url=f"{TEMPDIR_PREFIX}/compress/{schema}/file{ext}",
                is_archive=True,
                file_extension=ext,
            )

            # test for dir copy
            dir_copy_should_fail = schema in ("http", "https")
            dir_copy_fail_reason = "Copy from HTTP(S) dir is unsupported"
            test_configs.append(
                CopyTestConfig(
                    source=cloud_folder,
                    destination=local_compressed_archive,
                    compress_flag=True,
                    should_fail=dir_copy_should_fail,
                    fail_reason=dir_copy_fail_reason,
                )
            )
        if archive_types:
            # use the values for cloud_archive and local_archive
            # from last loop iteration
            test_configs.append(
                CopyTestConfig(
                    source=cloud_archive,
                    destination=local_archive,
                )
            )
        test_configs.append(
            CopyTestConfig(
                source=cloud_folder,
                destination=local_copy_folder,
                should_fail=dir_copy_should_fail,
                fail_reason=dir_copy_fail_reason,
            )
        )

    logger.info(f"Generated cloud to local tests: {test_configs}")
    return test_configs



import logging
from pathlib import Path

import pytest

from .conftest import CLIRunner


root_logger = logging.getLogger()
root_logger.setLevel(logging.INFO)


@pytest.mark.smoke
def test_init_aliases(cli_runner: CLIRunner) -> None:
    toml_path = Path(".neuro.toml")
    assert not toml_path.exists()

    result = cli_runner(["neuro-extras", "init-aliases"])
    assert result.returncode == 0, result
    assert "Added aliases to" in result.stdout, result

    assert toml_path.exists()



import logging
import os
import sys
import textwrap
import uuid
from pathlib import Path
from typing import Callable, ContextManager, Optional

import pytest
from tenacity import retry, stop_after_attempt, wait_random_exponential

from neuro_extras.const import EX_PLATFORMERROR

from .conftest import CLIRunner, Secret, gen_random_file


LOGGER = logging.getLogger(__name__)

NEURO_EXTRAS_PRESET: Optional[str] = os.environ.get("NEURO_EXTRAS_PRESET")

PRESET_ARG = ["--preset", NEURO_EXTRAS_PRESET] if NEURO_EXTRAS_PRESET else []


@pytest.mark.serial  # first we build the image, then we are trying to overwrite it
@pytest.mark.xfail
@pytest.mark.skipif(sys.platform == "win32", reason="kaniko does not work on Windows")
@pytest.mark.parametrize("overwrite", [True, False])
@retry(stop=stop_after_attempt(5), wait=wait_random_exponential(min=10, max=60))
def test_image_build_overwrite(
    cli_runner: CLIRunner,
    overwrite: bool,
    dockerhub_auth_secret: Secret,
) -> None:
    result = cli_runner(["neuro-extras", "init-aliases"])
    assert result.returncode == 0, result

    dockerfile_path = Path("nested/custom.Dockerfile")
    dockerfile_path.parent.mkdir(parents=True)

    random_file_to_disable_layer_caching = gen_random_file(dockerfile_path.parent)
    with open(dockerfile_path, "w") as f:
        f.write(
            textwrap.dedent(
                f"""\
                FROM ghcr.io/neuro-inc/alpine:latest
                ADD {random_file_to_disable_layer_caching} /tmp
                RUN echo !
                """
            )
        )

    python_version = f"{sys.version_info[0]}.{sys.version_info[1]}"
    img_uri_str = f"image:extras-e2e-overwrite:{sys.platform}-{python_version}-latest"
    build_command = [
        "neuro",
        "image-build",
        *PRESET_ARG,
        "-e",
        f"{dockerhub_auth_secret.name}=secret:{dockerhub_auth_secret.name}",
        "-f",
        str(dockerfile_path),
        ".",
        img_uri_str,
    ]
    if overwrite:
        build_command.insert(2, "-F")

    result = cli_runner(build_command)
    if overwrite:
        assert result.returncode == 0, result
    else:
        assert result.returncode == EX_PLATFORMERROR, result
    try:
        cli_runner(
            ["neuro", "image", "size", img_uri_str],
        )
    finally:
        # Only delete image after second run of the test
        if overwrite is False:
            # (A.K.) on GCP we get Illegal argument(s) ({"errors":
            # [{"code":"GOOGLE_MANIFEST_DANGLING_TAG",
            # "message":"Manifest is still referenced by tag: v1"}]})
            # cli_runner(["neuro", "image", "rm", img_uri_str])
            pass


@pytest.mark.xfail
@pytest.mark.skipif(sys.platform == "win32", reason="kaniko does not work on Windows")
@retry(stop=stop_after_attempt(5), wait=wait_random_exponential(min=10, max=60))
def test_ignored_files_are_not_copied(
    cli_runner: CLIRunner,
    dockerhub_auth_secret: Secret,
) -> None:
    result = cli_runner(["neuro-extras", "init-aliases"])
    assert result.returncode == 0, result

    dockerfile_path = Path("nested/custom.Dockerfile")
    dockerfile_path.parent.mkdir(parents=True)

    ignored_file = Path("this_file_should_not_be_added.txt")
    ignored_file.touch()
    Path(".neuroignore").write_text(f"{ignored_file}\n")

    random_file_to_disable_layer_caching = gen_random_file(dockerfile_path.parent)
    Path(dockerfile_path).write_text(
        textwrap.dedent(
            f"""\
            FROM ghcr.io/neuro-inc/alpine:latest
            ADD {random_file_to_disable_layer_caching} /tmp
            ADD . /
            RUN find / -name "*{random_file_to_disable_layer_caching.stem}*"
            RUN find / -name "*{ignored_file.stem}*"
            """
        )
    )

    img_uri_str = f"image:extras-e2e:{uuid.uuid4()}"

    cmd = [
        "neuro",
        "image-build",
        *PRESET_ARG,
        "-e",
        f"{dockerhub_auth_secret.name}=secret:{dockerhub_auth_secret.name}",
        "-f",
        str(dockerfile_path),
        ".",
        img_uri_str,
    ]
    result = cli_runner(
        cmd,
    )
    assert result.returncode == 0, result
    try:
        assert random_file_to_disable_layer_caching.name in result.stdout
        assert ignored_file.name not in result.stdout
    finally:
        # (A.K.) on GCP we get Illegal argument(s) ({"errors":
        # [{"code":"GOOGLE_MANIFEST_DANGLING_TAG",
        # "message":"Manifest is still referenced by tag: v1"}]})
        # cli_runner(["neuro", "image", "rm", img_uri_str])
        pass


@pytest.mark.serial
@pytest.mark.xfail
@pytest.mark.skipif(sys.platform == "win32", reason="kaniko does not work on Windows")
@retry(stop=stop_after_attempt(5), wait=wait_random_exponential(min=10, max=60))
def test_image_transfer(
    cli_runner: CLIRunner,
    switch_cluster: Callable[[str], ContextManager[None]],
    current_user: str,
    dockerhub_auth_secret: Secret,
    src_cluster: str,
    dst_cluster: str,
) -> None:
    # Note: we build src image on src_cluster and run transfer job in dst_cluster
    assert src_cluster != dst_cluster

    with switch_cluster(src_cluster):
        result = cli_runner(["neuro-extras", "init-aliases"])
        assert result.returncode == 0, result

        img_name = f"extras-e2e-image-copy:{str(uuid.uuid4())}"
        from_img = f"image:{img_name}"  # also, full src uri is supported
        to_img = f"image://{dst_cluster}/{current_user}/{img_name}"

        dockerfile_path = Path("nested/custom.Dockerfile")
        dockerfile_path.parent.mkdir(parents=True)

        random_file_to_disable_layer_caching = gen_random_file(dockerfile_path.parent)
        with open(dockerfile_path, "w") as f:
            f.write(
                textwrap.dedent(
                    f"""\
                    FROM ghcr.io/neuro-inc/alpine:latest
                    ADD {random_file_to_disable_layer_caching} /tmp
                    RUN echo !
                    """
                )
            )

        cmd = [
            "neuro",
            "image-build",
            *PRESET_ARG,
            "-e",
            f"{dockerhub_auth_secret.name}=secret:{dockerhub_auth_secret.name}",
            "-f",
            str(dockerfile_path),
            ".",
            from_img,
        ]
        result = cli_runner(cmd)
        assert result.returncode == 0, result

        try:
            cli_runner(
                ["neuro", "image", "size", from_img],
            )
            result = cli_runner(["neuro", "image-transfer", from_img, to_img])
            assert result.returncode == 0, result
            cli_runner(
                ["neuro", "image", "size", to_img],
            )
        finally:
            # (A.K.) on GCP we get Illegal argument(s) ({"errors":
            # [{"code":"GOOGLE_MANIFEST_DANGLING_TAG",
            # "message":"Manifest is still referenced by tag: v1"}]})
            # cli_runner(["neuro", "image", "rm", from_img])
            # cli_runner(["neuro", "image", "rm", to_img])
            pass


@pytest.mark.xfail
@pytest.mark.skipif(sys.platform == "win32", reason="kaniko does not work on Windows")
@pytest.mark.parametrize("img_repo_name", ["ne-test-public", "ne-test-private"])
@pytest.mark.parametrize("img_tag", ["", ":latest", ":v1.0.0"])
@retry(stop=stop_after_attempt(5), wait=wait_random_exponential(min=10, max=60))
def test_external_image_build(
    cli_runner: CLIRunner,
    dockerhub_auth_secret: Secret,
    img_repo_name: str,
    img_tag: str,
) -> None:
    dckrhb_uname = os.environ["DOCKER_CI_USERNAME"]

    result = cli_runner(["neuro-extras", "init-aliases"])
    assert result.returncode == 0, result

    dockerfile_path = Path("nested/custom.Dockerfile")
    dockerfile_path.parent.mkdir(parents=True)

    random_file_to_disable_layer_caching = gen_random_file(dockerfile_path.parent)
    with open(dockerfile_path, "w") as f:
        f.write(
            textwrap.dedent(
                f"""\
                FROM ghcr.io/neuro-inc/alpine:latest
                ADD {random_file_to_disable_layer_caching} /tmp
                RUN echo !
                """
            )
        )
    img_uri_str = f"{dckrhb_uname}/{img_repo_name}{img_tag}"
    build_command = [
        "neuro",
        "image-build",
        *PRESET_ARG,
        "-e",
        f"{dockerhub_auth_secret.name}=secret:{dockerhub_auth_secret.name}",
        "-f",
        str(dockerfile_path),
        ".",
        img_uri_str,
    ]
    result = cli_runner(
        build_command,
    )
    if f"Successfully built {img_uri_str}" not in result.stdout:
        LOGGER.warning(result.stdout)
        raise AssertionError("Successfully built message was not found.")


@pytest.mark.skipif(
    sys.platform == "win32", reason="docker is not installed on Windows nodes"
)
@pytest.mark.skipif(
    sys.platform == "darwin", reason="docker is not installed on Mac nodes"
)
@pytest.mark.xfail
@retry(stop=stop_after_attempt(5), wait=wait_random_exponential(min=10, max=60))
def test_image_local_build(cli_runner: CLIRunner) -> None:
    dockerfile_path = Path("nested/custom.Dockerfile")
    dockerfile_path.parent.mkdir(parents=True)

    if sys.platform == "win32":
        base_image = "hello-world"
    else:
        base_image = "ghcr.io/neuro-inc/alpine:latest"

    with open(dockerfile_path, "w") as f:
        f.write(
            textwrap.dedent(
                f"""\
                    FROM {base_image}

                    ARG CLOUD_SDK_VERSION=347.0.0
                    ENV CLOUD_SDK_VERSION=$CLOUD_SDK_VERSION

                    RUN echo sdk=$CLOUD_SDK_VERSION
                """
            )
        )

    tag = str(uuid.uuid4())
    img_uri_str = f"image:extras-e2e:{tag}"
    cmd = [
        "neuro-extras",
        "image",
        "local-build",
        "--verbose",
        "true",
        "-f",
        str(dockerfile_path),
        "--build-arg",
        f"CLOUD_SDK_VERSION=arg-{tag}",
        ".",
        img_uri_str,
    ]
    result = cli_runner(cmd)
    assert result.returncode == 0, result
    try:
        if sys.platform == "win32":
            arg_string = f" --build-arg CLOUD_SDK_VERSION=arg-{tag}"
        else:
            arg_string = f"sdk=arg-{tag}"
        assert arg_string in result.stderr or arg_string in result.stdout
    finally:
        # (A.K.) on GCP we get Illegal argument(s) ({"errors":
        # [{"code":"GOOGLE_MANIFEST_DANGLING_TAG",
        # "message":"Manifest is still referenced by tag: v1"}]})
        # cli_runner(["neuro", "image", "rm", img_uri_str])
        pass



import asyncio
import logging
import os
import re
import subprocess
import uuid
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from subprocess import CompletedProcess
from tempfile import TemporaryDirectory
from typing import (
    AsyncIterator,
    Callable,
    ContextManager,
    Dict,
    Iterator,
    List,
    Optional,
    Protocol,
    Union,
)

import neuro_sdk  # NOTE: don't use async test functions (issue #129)
import pytest
from tenacity import retry, stop_after_attempt, stop_after_delay

from neuro_extras.common import NEURO_EXTRAS_IMAGE
from neuro_extras.config import _build_registy_auth
from neuro_extras.image_builder import KANIKO_AUTH_PREFIX
from neuro_extras.utils import setup_child_watcher


DISK_PREFIX = "<DISK_PREFIX>"
TEMPDIR_PREFIX = "<TEMPDIR_PREFIX>"

UUID4_PATTERN = r"[0-9a-f]{8}-[0-9a-f]{4}-[4][0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}"
DISK_ID_PATTERN = rf"disk-{UUID4_PATTERN}"
DISK_ID_REGEX = re.compile(DISK_ID_PATTERN)

logger = logging.getLogger(__name__)

TEST_DATA_COPY_LOCAL_TO_LOCAL = True
TEST_DATA_COPY_CLOUD_TO_LOCAL = True
TEST_DATA_COPY_LOCAL_TO_CLOUD = True
TEST_DATA_COPY_CLOUD_TO_PLATFORM = True
TEST_DATA_COPY_PLATFORM_TO_CLOUD = True

CLOUD_SOURCE_PREFIXES: Dict[str, str] = {
    "gs": "gs://mlops-ci-e2e/assets/data",
    # "s3": "s3://because-clear-taken-cotton/assets/data",
    # "azure+https": "azure+https://neuromlops.blob.core.windows.net/cookiecutter-e2e/assets/data",  # noqa: E501
    "http": "http://because-clear-taken-cotton.s3.amazonaws.com/assets/data",
    "https": "https://because-clear-taken-cotton.s3.amazonaws.com/assets/data",
}

CLOUD_DESTINATION_PREFIXES: Dict[str, str] = {
    # "s3": "s3://because-clear-taken-cotton/data_cp",
    "gs": "gs://mlops-ci-e2e/data_cp",
    # "azure+https": "azure+https://neuromlops.blob.core.windows.net/cookiecutter-e2e/data_cp",  # noqa: E501
    "http": "http://because-clear-taken-cotton.s3.amazonaws.com/data_cp",
    "https": "https://because-clear-taken-cotton.s3.amazonaws.com/data_cp",
}

PLATFORM_SOURCE_PREFIXES: Dict[str, str] = {
    # neuro mkdir -p storage:e2e/assets/data
    # neuro cp -rT tests/assets/data storage:e2e/assets/data
    "storage": "storage:e2e/assets/data",
    # neuro disk create --name extras-e2e --timeout-unused 1000d 100M
    # neuro run -v storage:e2e/assets/data:/storage -v disk:extras-e2e:/disk alpine -- cp -rT /storage /disk/assets/data # noqa: E501
    "disk": f"disk:extras-e2e/assets/data",
}

PLATFORM_DESTINATION_PREFIXES: Dict[str, str] = {
    # neuro storage mkdir storage:e2e/data_cp
    "storage": "storage:e2e/data_cp",
    "disk": f"{DISK_PREFIX}/data_cp",
}

SRC_CLUSTER_ENV_VAR = "NEURO_CLUSTER"
DST_CLUSTER_ENV_VAR = "NEURO_CLUSTER_SECONDARY"


class CLIRunner(Protocol):
    def __call__(
        self, args: List[str], enable_retry: bool = False
    ) -> "CompletedProcess[str]":
        ...


def get_tested_archive_types() -> List[str]:
    """Get tested archive types

    If PYTEST_DATA_COPY_ARCHIVE_TYPES is set,
    returns its value, split on `,`.
    Otherwise, returns default archive types.

    See `neuro_extras.data.archive.ArchiveType.get_extension_mapping()`
    for supported extensions.
    """
    env = os.environ.get("PYTEST_DATA_COPY_ARCHIVE_TYPES")
    if env:
        return env.split(",")
    else:
        return [".tar.gz"]


setup_child_watcher()


@dataclass
class Secret:
    name: str
    value: str

    def __repr__(self) -> str:
        return f"Secret(name='{self.name}', value='HIDDEN!'"


def generate_random_secret(name_prefix: str = "secret") -> Secret:
    return Secret(
        name=f"{name_prefix}-{uuid.uuid4().hex[:8]}",
        value=str(uuid.uuid4()),
    )


@pytest.fixture
def temp_random_secret(cli_runner: CLIRunner) -> Iterator[Secret]:
    secret = generate_random_secret()
    try:
        yield secret
    finally:
        cli_runner(["neuro", "secret", "rm", secret.name])


def gen_random_file(location: Union[str, Path], name: Optional[str] = None) -> Path:
    location = Path(location)
    location.mkdir(parents=True, exist_ok=True)
    name = name or f"file-{uuid.uuid4().hex[:8]}.txt"
    file = location / name
    file.write_text(str(uuid.uuid4()))
    return file


@pytest.fixture(scope="session", autouse=True)
def print_neuro_extras_image() -> None:
    logger.warning(f"Using neuro-extras image: '{NEURO_EXTRAS_IMAGE}'")


async def _async_get_bare_client() -> neuro_sdk.Client:
    """Return uninitialized neuro client."""
    return await neuro_sdk.get()


@pytest.fixture
def _neuro_client() -> Iterator[neuro_sdk.Client]:
    # Note: because of issue #129 we can't use async methods of the client,
    # therefore this fixture is private
    client = asyncio.run(_async_get_bare_client())
    try:
        yield asyncio.run(client.__aenter__())
    finally:
        asyncio.run(client.__aexit__())  # it doesn't use arguments


@pytest.fixture
def current_user(_neuro_client: neuro_sdk.Client) -> str:
    return _neuro_client.username


@pytest.fixture
def switch_cluster(
    _neuro_client: neuro_sdk.Client,
) -> Callable[[str], ContextManager[None]]:
    @contextmanager
    def _f(cluster: str) -> Iterator[None]:
        orig_cluster = _neuro_client.config.cluster_name
        try:
            logger.info(f"Temporary cluster switch: {orig_cluster} -> {cluster}")
            asyncio.run(_neuro_client.config.switch_cluster(cluster))
            yield
        finally:
            logger.info(f"Switch back cluster: {cluster} -> {orig_cluster}")
            try:
                asyncio.run(_neuro_client.config.switch_cluster(orig_cluster))
            except Exception as e:
                logger.error(
                    f"Could not switch back to cluster '{orig_cluster}': {e}. "
                    f"Please run manually: 'neuro config switch-cluster {orig_cluster}'"
                )

    return _f


@pytest.fixture
async def dockerhub_auth_secret() -> AsyncIterator[Secret]:
    async with neuro_sdk.get() as neuro_client:
        secret_name = f"{KANIKO_AUTH_PREFIX}_{uuid.uuid4().hex}"
        auth_data = _build_registy_auth(
            # Why not v2: https://github.com/GoogleContainerTools/kaniko/pull/1209
            registry_uri="https://index.docker.io/v1/",
            username=os.environ["DOCKER_CI_USERNAME"],
            password=os.environ["DOCKER_CI_TOKEN"],
        )
        secret = Secret(secret_name, auth_data)
        try:
            await neuro_client.secrets.add(secret_name, auth_data.encode())
            logger.debug(f"Created test secret: {secret}")
            yield secret
        finally:
            await neuro_client.secrets.rm(secret_name)


@pytest.fixture
def project_dir() -> Iterator[Path]:
    with TemporaryDirectory() as cwd_str:
        old_cwd = Path.cwd()
        cwd = Path(cwd_str)
        os.chdir(cwd)
        try:
            yield cwd
        finally:
            os.chdir(old_cwd)


@retry(stop=stop_after_attempt(3) | stop_after_delay(5 * 10))
def run_cli(args: List[str]) -> "CompletedProcess[str]":
    proc = subprocess.run(
        args,
        check=False,
        capture_output=True,
        text=True,
    )
    if proc.returncode:
        logger.warning(f"Got '{proc.returncode}' for '{' '.join(args)}'")
    logger.info(f"stdout of {args}: {proc.stdout}")
    if proc.stderr:
        logger.warning(f"stderr of {args}: {proc.stderr}")
    return proc


@pytest.fixture
def cli_runner(project_dir: Path) -> CLIRunner:
    return run_cli  # type: ignore


@pytest.fixture
def args_data_cp_from_cloud(cli_runner: CLIRunner) -> Callable[..., List[str]]:
    def _f(
        bucket: str,
        src: str,
        dst: str,
        extract: bool,
        compress: bool,
        use_temp_dir: bool,
    ) -> List[str]:
        args = ["neuro-extras", "data", "cp", src, dst]
        if (
            src.startswith("storage:")
            or dst.startswith("storage:")
            or src.startswith("disk:")
            or dst.startswith("disk:")
        ):
            if bucket.startswith("gs://"):
                args.extend(
                    [
                        "-v",
                        "secret:neuro-extras-gcp:/gcp-creds.txt",
                        "-e",
                        "GOOGLE_APPLICATION_CREDENTIALS=/gcp-creds.txt",
                    ]
                )
            elif bucket.startswith("s3://"):
                args.extend(
                    [
                        "-v",
                        "secret:neuro-extras-aws:/aws-creds.txt",
                        "-e",
                        "AWS_CONFIG_FILE=/aws-creds.txt",
                    ]
                )
            elif bucket.startswith("azure+https://"):
                args.extend(
                    [
                        "-e",
                        "AZURE_SAS_TOKEN=secret:azure_sas_token",
                    ]
                )
            elif bucket.startswith("https://") or bucket.startswith("http://"):
                # No additional arguments required
                pass
            else:
                raise NotImplementedError(bucket)
        if extract:
            args.append("-x")
        if compress:
            args.append("-c")
        if use_temp_dir:
            args.append("-t")
        logger.info("args = %s", args)
        return args

    return _f


@pytest.fixture
def disk(cli_runner: CLIRunner) -> Iterator[str]:
    """Provide temporary disk, which will be removed upon test end

    WARNING: when used with pytest-xdist,
    disk will be created one per each worker!
    """
    # Create disk
    res = cli_runner(["neuro", "disk", "create", "100M"])
    assert res.returncode == 0, res
    disk_id = None
    try:
        output_lines = "\n".join(res.stdout.splitlines())

        search = DISK_ID_REGEX.search(output_lines)
        if search:
            disk_id = search.group()
        else:
            raise Exception("Can't find disk ID in neuro output: \n" + res.stdout)
        logger.info(f"Created disk {disk_id}")
        yield f"disk:{disk_id}"

    finally:
        logger.info(f"Removing disk {disk_id}")
        try:
            # Delete disk
            if disk_id is not None:
                res = cli_runner(["neuro", "disk", "rm", disk_id])
                assert res.returncode == 0, res
        except BaseException as e:
            logger.warning(f"Finalization error: {e}")


@pytest.fixture(scope="session")
def src_cluster() -> Iterator[str]:
    res = os.environ.get(SRC_CLUSTER_ENV_VAR)
    if not res:
        raise ValueError(f"'{SRC_CLUSTER_ENV_VAR}' env var is missing")
    yield res


@pytest.fixture(scope="session")
def dst_cluster() -> Iterator[str]:
    res = os.environ.get(DST_CLUSTER_ENV_VAR)
    if not res:
        pytest.skip(
            f"{DST_CLUSTER_ENV_VAR} env var"
            " indicating destination cluster is missing, skipping test"
        )
    yield res



import uuid
from typing import List

from ..conftest import (
    CLOUD_DESTINATION_PREFIXES,
    PLATFORM_SOURCE_PREFIXES,
    get_tested_archive_types,
)
from .resources import CopyTestConfig, DataTestResource


def generate_platform_to_cloud_copy_configs() -> List[CopyTestConfig]:
    test_configs: List[CopyTestConfig] = []
    archive_types = get_tested_archive_types()
    run_uuid = uuid.uuid4().hex
    for source_schema, source_prefix in PLATFORM_SOURCE_PREFIXES.items():
        source_archives = [
            DataTestResource(
                schema=source_schema,
                url=f"{source_prefix}/file{ext}",
                is_archive=True,
                file_extension=ext,
            )
            for ext in archive_types
        ]
        source_folder = DataTestResource(
            schema=source_schema,
            url=f"{source_prefix}/",
            is_archive=False,
            file_extension=None,
        )
        for schema, prefix in CLOUD_DESTINATION_PREFIXES.items():
            # tests for copy of local folder
            cloud_folder = DataTestResource(
                schema=schema,
                url=f"{prefix}/{run_uuid}/copy/",
                is_archive=False,
                file_extension=None,
            )
            test_configs.append(
                CopyTestConfig(source=source_folder, destination=cloud_folder)
            )

            # tests for compression of local folder
            for archive_type in archive_types:
                cloud_archive = DataTestResource(
                    schema=schema,
                    url=f"{prefix}/{run_uuid}/compress/file{archive_type}",
                    is_archive=True,
                    file_extension=archive_type,
                )
                # gzip does not work properly with folders
                # tar should be used instead
                compression_should_fail = archive_type == ".gz"
                compression_fail_reason = "gzip does not support folders properly"
                test_configs.append(
                    CopyTestConfig(
                        source=source_folder,
                        destination=cloud_archive,
                        compress_flag=True,
                        should_fail=compression_should_fail,
                        fail_reason=compression_fail_reason,
                    )
                )

            # tests for copy of local files
            for archive in source_archives:
                # test for extraction of archive
                cloud_extraction_folder = DataTestResource(
                    schema=schema,
                    url=f"{prefix}/{run_uuid}/extract/{archive.file_extension}/",
                    is_archive=False,
                    file_extension=None,
                )
                test_configs.append(
                    CopyTestConfig(
                        source=archive,
                        destination=cloud_extraction_folder,
                        extract_flag=True,
                    )
                )

                # test for file copy
                cloud_file = DataTestResource(
                    schema=schema,
                    url=f"{prefix}/{run_uuid}/copy/file{archive.file_extension}",
                    is_archive=True,
                    file_extension=archive.file_extension,
                )

                # test for skipping compression
                test_configs.append(
                    CopyTestConfig(
                        source=archive, destination=cloud_file, compress_flag=True
                    )
                )
            if source_archives:
                # use the values for source and destination files
                # from last loop iteration
                test_configs.append(
                    CopyTestConfig(source=archive, destination=cloud_file)
                )
    return test_configs



import asyncio
import base64
import json
import logging
import re
import shlex
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, AsyncIterator, Dict, List, Optional, Sequence, Tuple, Type

import click
import neuro_sdk
from neuro_cli.formatters.images import DockerImageProgress
from neuro_sdk._url_utils import _extract_path
from rich.console import Console
from yarl import URL


KANIKO_IMAGE_REF = "gcr.io/kaniko-project/executor"
KANIKO_IMAGE_TAG = "v1.20.0-debug"  # debug has busybox, which is needed for auth
KANIKO_AUTH_PREFIX = "NE_REGISTRY_AUTH"
KANIKO_DOCKER_CONFIG_PATH = "/kaniko/.docker/config.json"
KANIKO_AUTH_SCRIPT_PATH = "/kaniko/.docker/merge_docker_auths.sh"
KANIKO_CONTEXT_PATH = "/kaniko_context"
KANIKO_EXTRA_ENVS = ("container=docker",)
BUILDER_JOB_LIFESPAN = "4h"
BUILDER_JOB_SHEDULE_TIMEOUT = "20m"

MIN_BUILD_PRESET_CPU: float = 2
MIN_BUILD_PRESET_MEM: int = 4096

logger = logging.getLogger(__name__)


@dataclass
class DockerConfigAuth:
    registry: str
    username: str
    password: str = field(repr=False)

    @property
    def credentials(self) -> str:
        return base64.b64encode(f"{self.username}:{self.password}".encode()).decode()


@dataclass
class DockerConfig:
    auths: Sequence[DockerConfigAuth] = ()

    def to_primitive(self) -> Dict[str, Any]:
        return {
            "auths": {auth.registry: {"auth": auth.credentials} for auth in self.auths}
        }


async def create_docker_config_auth(
    client_config: neuro_sdk.Config,
) -> DockerConfigAuth:
    # retrieve registry hostname with optional port
    url = client_config.registry_url
    assert url.host
    port = f":{url.explicit_port}" if url.explicit_port else ""  # type: ignore
    registry_host = url.host + port
    auth = DockerConfigAuth(
        registry=registry_host,
        username=client_config.username,
        password=await client_config.token(),
    )
    return auth


class ImageBuilder(ABC):
    def __init__(
        self,
        client: neuro_sdk.Client,
        extra_registry_auths: Sequence[DockerConfigAuth] = (),
        verbose: bool = False,
    ) -> None:
        """
            Builds and pushes docker image to the platform.
            By default, build  happens on the platform, using Kaniko tool,
            unless --local is specified.

        Args:
            client (neuro_sdk.Client): instance of neuro-sdk client,
                authenticated to the destination cluster
            extra_registry_auths (Sequence[DockerConfigAuth], optional):
                Sequence of extra docker container registry auth credits,
                useful if base image(s) hidden under private registry(es).
                Defaults to ().
            verbose (bool, optional): Whether to set Kaniko's verbosity to DEBUG.
                Defaults to False.
        """
        self._client = client
        self._extra_registry_auths = list(extra_registry_auths)
        self._verbose = verbose

    def _generate_build_uri(self, project_name: str) -> URL:
        return self._client.parse.normalize_uri(
            URL(f"storage:/{project_name}/.builds/{uuid.uuid4()}"),
        )

    async def create_docker_config(self) -> DockerConfig:
        dst_reg_auth = await create_docker_config_auth(self._client.config)
        return DockerConfig(auths=[dst_reg_auth] + self._extra_registry_auths)

    async def save_docker_config(self, docker_config: DockerConfig, uri: URL) -> None:
        async def _gen() -> AsyncIterator[bytes]:
            yield json.dumps(docker_config.to_primitive()).encode()

        await self._client.storage.create(uri, _gen())

    def parse_image_ref(self, image_uri_str: str) -> str:
        image = self._client.parse.remote_image(image_uri_str)
        return re.sub(r"^http[s]?://", "", image.as_docker_url())

    @abstractmethod
    async def build(
        self,
        dockerfile_path: Path,
        context_uri: URL,
        image: neuro_sdk.RemoteImage,
        use_cache: bool,
        build_args: Tuple[str, ...],
        volumes: Tuple[str, ...],
        envs: Tuple[str, ...],
        job_preset: Optional[str],
        build_tags: Tuple[str, ...],
        project_name: str,
        extra_kaniko_args: Optional[str],
    ) -> int:
        pass

    @staticmethod
    def get(local: bool) -> Type["ImageBuilder"]:
        if local:
            return LocalImageBuilder
        else:
            return RemoteImageBuilder

    async def _execute_subprocess(self, command: Sequence[str]) -> int:
        logger.debug("Executing subprocess: %s", " ".join(command))
        subprocess = await asyncio.create_subprocess_exec(*command)
        return await subprocess.wait()


class LocalImageBuilder(ImageBuilder):
    async def build(
        self,
        dockerfile_path: Path,
        context_uri: URL,
        image: neuro_sdk.RemoteImage,
        use_cache: bool,
        build_args: Tuple[str, ...],
        volumes: Tuple[str, ...],
        envs: Tuple[str, ...],
        job_preset: Optional[str],
        build_tags: Tuple[str, ...],
        project_name: str,
        extra_kaniko_args: Optional[str],
    ) -> int:
        logger.info(f"Building the image {image}")
        logger.info(f"Using {context_uri} as the build context")
        if extra_kaniko_args:
            logger.warning(
                "Extra kaniko args are not supported for local builds. "
                "They will be ignored."
            )

        docker_build_args = []

        for arg in build_args:
            docker_build_args.append(f"--build-arg {arg}")

        build_command = [
            "docker",
            "build",
            f"--tag={image.as_docker_url()}",
            f"--file={dockerfile_path}",
        ]
        if not self._verbose:
            build_command.append("--quiet")
        if len(docker_build_args) > 0:
            build_command.append(" ".join(docker_build_args))
        build_command.append(str(_extract_path(context_uri)))

        ex_code = await self._execute_subprocess(build_command)
        if ex_code != 0:
            return ex_code
        return await self._push_image(image)

    async def _push_image(self, image: neuro_sdk.RemoteImage) -> int:
        logger.info(f"Pushing image to registry")
        console = Console()
        progress = DockerImageProgress.create(console=console, quiet=not self._verbose)
        local_image = self._client.parse.local_image(image.as_docker_url())
        try:
            await self._client.images.push(local_image, image, progress=progress)
            logger.info(
                f"Pushed {image.as_docker_url()} to the platform registry as {image}"
            )
        except Exception as e:
            logger.exception("Image push failed.")
            logger.info(
                f"You may try to repeat the push process by running "
                f"'neuro image push {local_image} {image}'"
            )
            raise e
        return 0


class RemoteImageBuilder(ImageBuilder):
    async def build(
        self,
        dockerfile_path: Path,
        context_uri: URL,
        image: neuro_sdk.RemoteImage,
        use_cache: bool,
        build_args: Tuple[str, ...],
        volumes: Tuple[str, ...],
        envs: Tuple[str, ...],
        job_preset: Optional[str],
        build_tags: Tuple[str, ...],
        project_name: str,
        extra_kaniko_args: Optional[str],
    ) -> int:
        # TODO: check if Dockerfile exists
        logger.info(f"Building the image {image}")
        logger.info(f"Using {context_uri} as the build context")

        # upload (if needed) build context and platform registry auth info
        build_uri = self._generate_build_uri(project_name)
        await self._client.storage.mkdir(build_uri, parents=True)
        if context_uri.scheme == "file":
            storage_context_uri = build_uri / "context"
            await self._upload_to_storage(context_uri, storage_context_uri)
            context_uri = storage_context_uri

        docker_config = await self.create_docker_config()
        docker_config_uri = build_uri / ".docker.config.json"
        logger.debug(f"Uploading {docker_config_uri}")
        await self.save_docker_config(docker_config, docker_config_uri)

        cache_image = neuro_sdk.RemoteImage(
            name="layer-cache/cache",
            project_name=project_name,
            registry=str(self._client.config.registry_url),
            cluster_name=self._client.cluster_name,
        )
        cache_repo = self.parse_image_ref(str(cache_image))
        cache_repo = re.sub(r":.*$", "", cache_repo)  # drop tag

        if any(KANIKO_AUTH_PREFIX in env for env in envs):
            # we have extra auth info.
            # in this case we cannot mount registry auth info at the default path
            # and should upload and configure 'merge_docker_auths' script to merge auths
            mnt_path = Path(KANIKO_DOCKER_CONFIG_PATH)
            mnt_path = mnt_path.with_name(f"{mnt_path.stem}_base{mnt_path.suffix}")
            docker_config_mnt = str(mnt_path)
            envs += (
                f"{KANIKO_AUTH_PREFIX}_BASE_{uuid.uuid4().hex[:8]}={docker_config_mnt}",
            )
            local_script = URL(
                (Path(__file__).parent / "assets" / "merge_docker_auths.sh").as_uri()
            )
            remote_script = build_uri / "merge_docker_auths.sh"
            await self._client.storage.upload_file(local_script, remote_script)
            volumes += (f"{remote_script}:{KANIKO_AUTH_SCRIPT_PATH}:ro",)
            job_entrypoint_overwrite = [
                f"sh {KANIKO_AUTH_SCRIPT_PATH}",
                "&&",
                "executor",
                # Kaniko args will be added below
            ]
        else:
            docker_config_mnt = str(KANIKO_DOCKER_CONFIG_PATH)
            job_entrypoint_overwrite = []

        # mount build context and platform registry auth info
        volumes += (
            f"{docker_config_uri}:{docker_config_mnt}:ro",
            # context dir cannot be R/O if we want to mount secrets there
            f"{context_uri}:{KANIKO_CONTEXT_PATH}:rw",
        )
        build_tags += (f"kaniko-builds-image:{image}",)
        kaniko_args = [
            f"--context={KANIKO_CONTEXT_PATH}",
            f"--dockerfile={KANIKO_CONTEXT_PATH}/{dockerfile_path.as_posix()}",
            f"--destination={image.as_docker_url(with_scheme=False)}",
            f"--cache={'true' if use_cache else 'false'}",
            f"--cache-repo={cache_repo}",
            f"--verbosity={'debug' if self._verbose else 'info'}",
            "--image-fs-extract-retry=1",
            "--push-retry=3",
            "--use-new-run=true",
            "--snapshot-mode=redo",
        ]

        for arg in build_args:
            kaniko_args.append(f"--build-arg {arg}")
        # env vars (which might be platform secrets too) are passed as build args
        env_parsed = self._client.parse.envs(envs)
        for arg in list(env_parsed.env) + list(env_parsed.secret_env):
            if KANIKO_AUTH_PREFIX not in arg:
                kaniko_args.append(f"--build-arg {arg}")

        kaniko_args = self._add_extra_kaniko_args(kaniko_args, extra_kaniko_args)

        build_command = [
            "neuro",
            "--disable-pypi-version-check",
            "job",
            "run",
            f"--life-span={BUILDER_JOB_LIFESPAN}",
            f"--schedule-timeout={BUILDER_JOB_SHEDULE_TIMEOUT}",
            f"--project={project_name}",
        ]
        if job_preset:
            build_command.append(f"--preset={job_preset}")
        for build_tag in build_tags:
            build_command.append(f"--tag={build_tag}")
        for volume in volumes:
            build_command.append(f"--volume={volume}")
        for env in envs:
            build_command.append(f"--env={env}")
        envs_keys = [e.split("=")[0] for e in envs]
        for extra_env in KANIKO_EXTRA_ENVS:
            if extra_env.split("=")[0] in envs_keys:
                logger.warning(
                    f"Cannot overwite env {extra_env}: already present. "
                    "Consider removing this environment variable from your config, "
                    "otherwise, the build might fail."
                )
            else:
                build_command.append(f"--env={extra_env}")

        kaniko_args_str = " ".join(kaniko_args)
        if job_entrypoint_overwrite:
            job_entrypoint_overwrite.append(kaniko_args_str)
            build_command.append("--entrypoint")
            build_command.append(
                f"sh -c {shlex.quote(' '.join(job_entrypoint_overwrite))}"
            )
            build_command.append(f"{KANIKO_IMAGE_REF}:{KANIKO_IMAGE_TAG}")
        else:
            build_command.append(f"{KANIKO_IMAGE_REF}:{KANIKO_IMAGE_TAG}")
            build_command.append("--")
            build_command.append(kaniko_args_str)

        # TODO: remove context after the build is finished?
        return await self._execute_subprocess(build_command)

    async def _upload_to_storage(self, local_url: URL, remote_url: URL) -> None:
        logger.info(f"Uploading {local_url} to {remote_url}")
        command = [
            "neuro",
            "--disable-pypi-version-check",
            "cp",
            "--recursive",
            str(local_url),
            str(remote_url),
        ]
        return_code = await self._execute_subprocess(command)
        if return_code != 0:
            raise click.ClickException("Uploading build context failed!")

    def _add_extra_kaniko_args(
        self, kaniko_args: List[str], extra_kaniko_args: Optional[str]
    ) -> List[str]:
        if not extra_kaniko_args:
            return kaniko_args

        extra_args = shlex.split(extra_kaniko_args)
        kaniko_arg_keys = [arg.split("=")[0] for arg in kaniko_args]
        extra_args_keys = [arg.split("=")[0] for arg in extra_args]
        overlap = set(extra_args_keys) & set(kaniko_arg_keys)
        if not overlap:
            return kaniko_args + extra_args
        raise ValueError(
            f"Extra kaniko arguments {overlap} overlap with autogenerated arguments. "
            "Please remove them in order to proceed or contact the support team."
        )



import base64
import json
from unittest import mock

import pytest
import yaml

from .conftest import CLIRunner


@pytest.mark.smoke
def test_k8s_generate_secret(cli_runner: CLIRunner) -> None:
    result = cli_runner(["neuro-extras", "k8s", "generate-secret"])
    assert result.returncode == 0, result

    payload = yaml.safe_load(result.stdout)
    assert payload == {
        "apiVersion": "v1",
        "kind": "Secret",
        "metadata": {"name": "neuro"},
        "type": "Opaque",
        "data": mock.ANY,
    }
    assert payload["data"]["db"]


def test_k8s_generate_secret_custom_name(cli_runner: CLIRunner) -> None:
    result = cli_runner(["neuro-extras", "k8s", "generate-secret", "--name", "test"])
    assert result.returncode == 0, result

    payload = yaml.safe_load(result.stdout)
    assert payload == {
        "apiVersion": "v1",
        "kind": "Secret",
        "metadata": {"name": "test"},
        "type": "Opaque",
        "data": mock.ANY,
    }
    assert payload["data"]["db"]


def test_k8s_generate_registry_secret(cli_runner: CLIRunner) -> None:
    result = cli_runner(["neuro-extras", "k8s", "generate-registry-secret"])
    assert result.returncode == 0, result

    payload = yaml.safe_load(result.stdout)
    assert payload == {
        "apiVersion": "v1",
        "kind": "Secret",
        "metadata": {"name": "neuro-registry"},
        "type": "kubernetes.io/dockerconfigjson",
        "data": {".dockerconfigjson": mock.ANY},
    }
    docker_config_payload = json.loads(
        base64.b64decode(payload["data"][".dockerconfigjson"])
    )
    assert docker_config_payload == {"auths": mock.ANY}


def test_k8s_generate_registry_secret_custom_name(cli_runner: CLIRunner) -> None:
    result = cli_runner(
        ["neuro-extras", "k8s", "generate-registry-secret", "--name", "test"]
    )
    assert result.returncode == 0, result

    payload = yaml.safe_load(result.stdout)
    assert payload == {
        "apiVersion": "v1",
        "kind": "Secret",
        "metadata": {"name": "test"},
        "type": "kubernetes.io/dockerconfigjson",
        "data": {".dockerconfigjson": mock.ANY},
    }
